/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.ui.view.fx;

import static javafx.scene.paint.Color.BLUE;
import static javafx.scene.paint.Color.DARKGRAY;
import static javafx.scene.paint.Color.GREEN;
import static javafx.scene.paint.Color.LIGHTGRAY;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_10PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_12PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_14PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_8PT;
import static org.fortiss.tooling.common.ui.javafx.style.LineStyle.SOLID_BLACK_1PT;
import static org.fortiss.tooling.kernel.ui.util.SelectionUtils.checkAndPickFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.text.DecimalFormat;
import java.time.format.DateTimeFormatter;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.fortiss.tooling.common.ui.javafx.layout.CompositeFXControllerBase;
import org.fortiss.tooling.common.ui.javafx.style.FillStyle;
import org.fortiss.tooling.common.ui.javafx.style.FontStyle;
import org.fortiss.tooling.common.ui.javafx.style.LineStyle;
import org.fortiss.tooling.ext.quality.MetricUpdateListener;
import org.fortiss.tooling.ext.quality.data.MetricDataManager;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;
import org.fortiss.tooling.ext.quality.service.ModelQualityService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.spiderchart.control.SpiderChartViewer;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.DoubleAxis;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.AxisStyle;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.DataSeriesStyle;
import org.fortiss.tooling.spiderchart.style.LegendStyle;

import javafx.application.Platform;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.chart.PieChart;
import javafx.scene.chart.ScatterChart;
import javafx.scene.chart.XYChart;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.Label;
import javafx.scene.control.SplitPane;
import javafx.scene.layout.BorderPane;
import javafx.scene.text.Text;

/**
 * FX Controller for the Metrics view.
 *
 * @author groh
 */

@SuppressWarnings("unchecked")
public class ModelQualityFXController extends CompositeFXControllerBase<SplitPane, Node>
		implements ISelectionListener, MetricUpdateListener {

	/** ChoiceBox for selecting the metric which shall be displayed. */
	@FXML
	private ChoiceBox<MetricKey> choiceBox;

	/** The bottom {@link BorderPane} for the spider chart. */
	@FXML
	private BorderPane borderPane;

	/** PieChart displaying the metrics. */
	@FXML
	private PieChart pieChart;

	/** A scatter chart for displaying metrics. */
	@FXML
	private ScatterChart<Number, Number> scatterChart;

	/** {@link Text} at the bottom of the view. */
	@FXML
	private Text bottomText;

	/** Time format when displaying updating time. */
	private final DateTimeFormatter timeFormat = DateTimeFormatter.ofPattern("HH:mm:ss");

	/** {@inheritDoc} */
	@Override
	public String getFXMLLocation() {
		return "MetricsViewFx.fxml";
	}

	/** {@inheritDoc} */
	@Override
	public void initialize() {
		choiceBox.getItems().addAll(MetricKey.values());
	}

	/**
	 * The last {@link EObject} which the user has clicked. Might be null when no or an invalid
	 * Object is selected.
	 */
	private EObject lastSelectedElement;

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {

		// Invalidate last selection, so if no new selection is found the charts are cleared
		lastSelectedElement = null;

		ModelQualityService.getInstance();
		var manager = ModelQualityService.getInstance().getMetricDataManagerInstance();

		EObject selected = checkAndPickFirst(selection, EObject.class);

		if(selected == null) {
			updateCharts("Nothing selected");
			return;
		}

		IProjectRootElement currentRootElement =
				selected instanceof IProjectRootElement ? (IProjectRootElement)selected
						: getFirstParentWithType(selected, IProjectRootElement.class);

		if(currentRootElement == null) {
			updateCharts("Selection does not have a ProjectRoot");
			return;
		}

		var rootDataElement = manager.getDataRootElementMap().get(currentRootElement);

		if(rootDataElement == null) {
			bottomText.setText(
					"Save model or select \"Export Metrics\" in project options to generate metrics");
			updateCharts("No metrics available");
			return;
		}

		bottomText.setText("Metrics last updated: " +
				timeFormat.format(rootDataElement.getLastRefresh()) + ". Save model to refresh");
		lastSelectedElement = selected;
		updateCharts("Selection does not have metric information");
	}

	/**
	 * Is triggered when the choice box changes.
	 */
	public void onChoiceBoxChange() {
		updateCharts("Invalid selection or metrics not available");
	}

	/**
	 * Updates the chart in the metric view.
	 * 
	 * @param message
	 *            to display if metric is not available
	 */
	private void updateCharts(String message) {

		if(lastSelectedElement != null) {
			MetricDataManager manager =
					ModelQualityService.getInstance().getMetricDataManagerInstance();
			MetricTreeNode node = manager.getTreeNodeLookupTable().get(lastSelectedElement);

			if(node != null) {

				Set<MetricKey> validKeys = new HashSet<>();
				validKeys.addAll(node.getStoredDoubles().keySet());
				validKeys.addAll(node.getStoredIntegers().keySet());
				for(MetricTreeNode child : node.getChildren()) {
					validKeys.addAll(child.getStoredDoubles().keySet());
					validKeys.addAll(child.getStoredIntegers().keySet());
				}

				ObservableList<MetricKey> choices = choiceBox.getItems();
				choices.retainAll(validKeys);
				for(MetricKey key : validKeys) {
					if(!choices.contains(key)) {
						choices.add(key);
					}
				}

				if(choiceBox.getValue() == null && !choices.isEmpty()) {
					// Set choice box to first element in choices, if there are any choices and none
					// is currently selected
					choiceBox.setValue(choices.get(0));
				}

				if(!node.getChildren().isEmpty()) {

					if(choiceBox.getValue() != null) {
						SpiderChartViewer viewer = buildSpiderChart(node, choiceBox.getValue());
						borderPane.setCenter(viewer.getViewerPane());

						// Create PieChart and add data
						pieChart.getData().clear();
						pieChart.setLegendVisible(false);
						pieChart.setTitle("");
						DecimalFormat format = new DecimalFormat("0.#");

						for(var child : node.getChildren()) {
							Double value = child.getValueAsDouble(choiceBox.getValue());
							if(value != null) {
								PieChart.Data slice = new PieChart.Data(
										child.getName() + ": " + format.format(value), value);
								pieChart.getData().add(slice);
							}
						}
					}
					// scatterChart
					scatterChart.getXAxis().setLabel("NUMBER_OF_TOTAL_LEAF_ELEMENTS");
					scatterChart.getYAxis().setLabel("NUMBER_OF_TOTAL_CYCLOMATIC_COMPLEXITY");

					scatterChart.getData().clear();

					for(var child : node.getChildren()) {

						Integer leafs = child.getStoredIntegers()
								.get(MetricKey.NUMBER_OF_TOTAL_LEAF_ELEMENTS);
						Integer complexity = child.getStoredIntegers()
								.get(MetricKey.NUMBER_OF_TOTAL_CODE_CYCLOMATIC_COMPLEXITY);

						if(leafs != null && complexity != null) {
							// Add data to the chart
							XYChart.Series<Number, Number> dataSeries = new XYChart.Series<>();
							dataSeries.getData().add(new XYChart.Data<>(leafs, complexity));
							scatterChart.getData().add(dataSeries);
						}
					}

					// Return, otherwise the data will be cleared
					return;
				}
				message = "Selection does not have contained elements";
			} else {
				message = "Selection does not have metric information";
			}
		} else {
			// Nothing selected, so remove the bottom text
			bottomText.setText("");
		}
		borderPane.setCenter(new Label(message));
		pieChart.getData().clear();
	}

	/**
	 * @param node
	 *            of which the data will be visualized
	 * @param key
	 *            {@link MetricKey} of the data to visualize
	 * @return {@link SpiderChartViewer} with the matching metrics
	 */
	private static SpiderChartViewer buildSpiderChart(MetricTreeNode node, MetricKey key) {
		SpiderChart spiderChart = new SpiderChart();
		spiderChart.setLegendLabel("Legend");
		ChartStyle chartStyle = new ChartStyle(true, true, true);

		List<MetricTreeNode> children = node.getChildren();

		if(children.isEmpty()) {
			return null;
		}

		double maxval = children.stream().map(p -> p.getValueAsDouble(key)).filter(p -> p != null)
				.mapToDouble(p -> p).max().orElse(0.0);

		chartStyle.setUseIndividualAxisSegments(false);
		chartStyle.setTitleStyle(new FontStyle("Verdana", 14, BLUE.brighter()));

		DataSeries elementData = new DataSeries("Data");
		for(var child : node.getChildren()) {
			Double value = child.getValueAsDouble(key);
			String name = child.getName();
			if(value == null) {
				value = 0.0;
			}
			if(name == null) {
				name = "Impossible null";
			}
			DoubleAxis testing = new DoubleAxis(name, 0.0, maxval);

			AxisStyle aStyle3Segs = new AxisStyle(SOLID_BLACK_1PT, BLACK_VERDANA_14PT, 3,
					BLACK_VERDANA_8PT, new DecimalFormat("#.##"));
			chartStyle.setAxisStyle(testing, aStyle3Segs);

			spiderChart.addAxis(testing);
			elementData.setPoint(testing, value);
		}
		spiderChart.addData(elementData);

		LineStyle olive1pt = new LineStyle(GREEN);
		FillStyle oliveFill = new FillStyle(GREEN, 0.25);
		DataSeriesStyle iphoneStyle = new DataSeriesStyle(olive1pt, oliveFill, true, true,
				BLACK_VERDANA_10PT, 7, new DecimalFormat("#.#"));
		chartStyle.setDataSeriesStyle(elementData, iphoneStyle);

		LegendStyle legendStyle = new LegendStyle(false, 5, BLACK_VERDANA_12PT);
		chartStyle.setLegendStyle(legendStyle);

		chartStyle.setAxisSegments(4);

		chartStyle.setBackgroundFillStyle(new FillStyle(LIGHTGRAY));
		chartStyle.setBackgroundLineStyle(new LineStyle(DARKGRAY));

		SpiderChartViewer viewer = new SpiderChartViewer(spiderChart, chartStyle);
		return viewer;
	}

	/** {@inheritDoc} */
	@Override
	public void metricUpdate(IProjectRootElement updatedElement) {
		Platform.runLater(() -> {
			if(lastSelectedElement == null) {
				// If nothing is selected, there is nothing to update
				return;
			}

			var rootDataElement = ModelQualityService.getInstance().getMetricDataManagerInstance()
					.getDataRootElementMap().get(updatedElement);
			if(rootDataElement != null) {
				bottomText.setText("Metrics last updated: " +
						timeFormat.format(rootDataElement.getLastRefresh()) +
						". Save model to refresh");
			} else {
				bottomText.setText(
						"Save model or select \"Export Metrics\" in project options to generate metrics");
			}
			updateCharts("Selection does not have metric information");
		});
	}
}
