/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality;

import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * {@link IMetricProvider} to collect various metrics from an {@link IHierarchicElement}.
 * 
 * @author groh
 */
public class HierarchicElementProvider implements IMetricProvider<IHierarchicElement> {
	/** {@inheritDoc} */
	@Override
	public void collectMetrics(MetricTreeNode node, IHierarchicElement currentElement) {

		var integers = node.getStoredIntegers();
		// Only collect these metrics if the name is null, in which case they have not been
		// collected before
		// This is used in some cases to combine metrics from two different IHierarchicElement
		if(node.getName() == null) {

			// Metrics are always set to default value, as these metric are combined later
			boolean isElementCommentable = currentElement instanceof INamedCommentedElement;
			integers.put(MetricKey.NUMBER_OF_TOTAL_COMMENTABLE_ELEMENTS,
					isElementCommentable ? 1 : 0);
			String comment = isElementCommentable
					? ((INamedCommentedElement)currentElement).getComment() : null;
			integers.put(MetricKey.NUMBER_OF_TOTAL_COMMENTED_ELEMENTS,
					comment != null && comment != "" ? 1 : 0);
			if(isElementCommentable) {
				node.setName(((INamedCommentedElement)currentElement).getName());
			} else {
				node.setName("Unnamable Element");
			}

			integers.put(MetricKey.UNQIUE_ID, (currentElement instanceof IIdLabeled)
					? ((IIdLabeled)currentElement).getId() : -1);
		}

		var connectors = currentElement.getConnectors();
		integers.put(MetricKey.NUMBER_OF_CONNECTORS, connectors.size());

		integers.put(MetricKey.NUMBER_OF_CONTAINED_ELEMENTS,
				currentElement.getContainedElements().size());
		integers.put(MetricKey.NUMBER_OF_CONNECTIONS, currentElement.getConnections().size());

		// depth metrics
		integers.put(MetricKey.NUMBER_OF_TOTAL_CONNECTORS, connectors.size());
		var entry_connectors = EcoreUtils.pickInstanceOf(EntryConnectorBase.class, connectors);
		var exit_connectors = EcoreUtils.pickInstanceOf(ExitConnectorBase.class, connectors);
		integers.put(MetricKey.NUMBER_OF_TOTAL_ENTRY_CONNECTORS, entry_connectors.size());
		integers.put(MetricKey.NUMBER_OF_TOTAL_EXIT_CONNECTORS, exit_connectors.size());

		integers.put(MetricKey.NUMBER_OF_TOTAL_ELEMENTS, 1);

		integers.put(MetricKey.NUMBER_OF_TOTAL_LEAF_ELEMENTS,
				currentElement.getContainedElements().size() == 0 ? 1 : 0);
	}
}
