/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.service;

import static java.util.Collections.emptyList;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.ext.quality.GraphMetricsProvider;
import org.fortiss.tooling.ext.quality.IMetricProvider;
import org.fortiss.tooling.ext.quality.MetricUpdateListener;
import org.fortiss.tooling.ext.quality.data.DataRootElement;
import org.fortiss.tooling.ext.quality.data.MetricDataManager;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;
import org.fortiss.tooling.ext.quality.storage.CSVFileWriter;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation.ESeverity;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.internal.ConstraintCheckerService;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IMigrationService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.service.listener.IPersistencyServiceListener;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * This class implements the {@link IMigrationService} interface.
 * 
 * @author blaschke
 * @author groh
 */
public class ModelQualityService extends EObjectAwareServiceBase<IMetricProvider<? extends EObject>>
		implements IIntrospectiveKernelService, IModelQualityService, IPersistencyServiceListener {

	/** The singleton instance. */
	private static final ModelQualityService INSTANCE = new ModelQualityService();

	/** Returns singleton instance of the service. */
	public static ModelQualityService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.ext.quality.metricService";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "metricService";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "metricService";

	/** Instance of the MetricDataManager. */
	private MetricDataManager metricDataManagerInstance = new MetricDataManager();

	/** List of {@link MetricUpdateListener} which are called when the metrics are updated */
	private List<MetricUpdateListener> metricUpdateListeners = new ArrayList<>();

	/** Top-level elements queued to be processed by the metric collector. */
	private final Queue<ITopLevelElement> queuedProcessableElements =
			new ConcurrentLinkedQueue<ITopLevelElement>();

	/** Stores the metrics collector job. */
	private final Job metricCollectorJob = new Job("Collecting metrics ...") {

		@Override
		protected IStatus run(IProgressMonitor monitor) {

			do {
				if(monitor.isCanceled()) {
					return Status.CANCEL_STATUS;
				}

				ITopLevelElement toBeProcessed = queuedProcessableElements.poll();
				if(toBeProcessed == null) {
					return Status.OK_STATUS;
				}

				performMetricCollection(toBeProcessed);
			} while(true);
		}
	};

	/** {@inheritDoc} */
	@Override
	public void scheduleMetricCollection(ITopLevelElement element) {
		queuedProcessableElements.add(element);
		metricCollectorJob.schedule();
	}

	/** {@inheritDoc} */
	@Override
	public void registerMetricUpdateListener(MetricUpdateListener listener) {
		this.metricUpdateListeners.add(listener);
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
		IPersistencyService.getInstance().addTopLevelElementListener(this);
	}

	/** Registers the metric provider with the service. */
	@Override
	public <T extends EObject> void registerMetricProvider(IMetricProvider<T> provider,
			Class<T> modelElementClass) {
		addHandler(modelElementClass, provider);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the tracking of metrical development in modeling." +
				"\n the service outputs the current metric analysis of the model and stores" +
				"\n the data within a csv for later history analysis." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/**
	 * Collects all metrics for all elements in the provided {@link ITopLevelElement}.
	 * 
	 * @param topLvl
	 *            the {@link ITopLevelElement} on which the analysis should be performed
	 */
	private void performMetricCollection(ITopLevelElement topLevelElement) {
		// Sort root elements, as some metrics are saved into trees from other root elements
		// To avoid these problems, the elements have to be processed in a certain order

		// Currently the only element affected by this is the allocation table, which is
		// why push it to the end of the list
		List<IProjectRootElement> rootElements = new ArrayList<>();
		rootElements.addAll(EcoreUtils.getChildrenWithType(topLevelElement.getRootModelElement(),
				IProjectRootElement.class));

		String allocationTableClassName =
				"org.fortiss.af3.allocation.model.impl.AllocationTableCollectionImpl";

		// Sort with custom comparator which pushes the object back if it matches the above class
		// name
		Collections.sort(rootElements, (r1, r2) -> {
			if(allocationTableClassName.equals(r1.getClass().getName())) {
				return 1;
			} else if(allocationTableClassName.equals(r2.getClass().getName())) {
				return -1;
			}
			return 0;
		});
		Map<IProjectRootElement, DataRootElement> updatedElements = new HashMap<>();
		for(IProjectRootElement rootElement : rootElements) {

			MetricTreeNode rootNode = new MetricTreeNode();

			if(rootElement instanceof IHierarchicElement) {

				IHierarchicElement firstElement = (IHierarchicElement)rootElement;
				recursivlyCollectMetrics(rootNode, firstElement, 0);
			} else {
				// Search for provider on ProjectRootElement
				collectMetrics(rootNode, rootElement);
			}

			// Collect all violated constraints and group them by their respective source
			List<IConstraintViolation<? extends EObject>> contraintViolations =
					ConstraintCheckerService.getInstance()
							.performAllConstraintChecksRecursively(rootElement);

			// Add the violated constraints to the corresponding tree node
			for(Entry<? extends EObject, List<IConstraintViolation<? extends EObject>>> entry : contraintViolations
					.stream().collect(Collectors.groupingBy(IConstraintViolation::getSource))
					.entrySet()) {

				MetricTreeNode node =
						metricDataManagerInstance.getTreeNodeLookupTable().get(entry.getKey());

				if(node != null) {
					// Collect violations and store them in the error and warning metric
					List<IConstraintViolation<? extends EObject>> violations = entry.getValue();
					var integers = node.getStoredIntegers();
					integers.put(MetricKey.CONSTRAINT_VIOLATIONS_ERROR, (int)violations.stream()
							.filter(v -> v.getSeverity() == ESeverity.ERROR).count());
					integers.put(MetricKey.CONSTRAINT_VIOLATIONS_WARNING, (int)violations.stream()
							.filter(v -> v.getSeverity() == ESeverity.WARNING).count());
				}
			}
			DataRootElement dataRootElement =
					new DataRootElement(LocalDateTime.now(), topLevelElement, rootNode);
			metricDataManagerInstance.getDataRootElementMap().put(rootElement, dataRootElement);
			updatedElements.put(rootElement, dataRootElement);
			metricDataManagerInstance.getTreeNodeLookupTable().put(rootElement, rootNode);
			metricUpdateListeners.forEach(l -> l.metricUpdate(rootElement));
		}

		// Store all currently saved metrics to the csv
		CSVFileWriter.metricExtractionToCSV(updatedElements);
	}

	/**
	 * Collect metrics for the given {@link IHierarchicElement} and all its contained elements.
	 * 
	 * @param node
	 *            to which the data is collected
	 * @param currentElement
	 *            element to collect the data from
	 * @param recursionLevel
	 *            parameter which indicates the current recursion level, used for metric
	 */
	private void recursivlyCollectMetrics(MetricTreeNode node, IHierarchicElement currentElement,
			int recursionLevel) {

		var manager = metricDataManagerInstance;

		collectMetrics(node, currentElement);
		node.getStoredIntegers().put(MetricKey.NESTING_LEVEL, recursionLevel);

		// Check if any of the specifications is an IHierarchicElement
		// This is for example the case with an StateAutomaton
		Optional<IHierarchicElement> hierarchicElementOptional = currentElement.getSpecifications()
				.stream().filter(s -> s instanceof IHierarchicElement)
				.map(s -> (IHierarchicElement)s).findAny();
		if(hierarchicElementOptional.isPresent()) {
			// Do not collect statistics for the current element, but instead collect them for the
			// specification
			IHierarchicElement specificationElement = hierarchicElementOptional.get();
			EList<IHierarchicElement> containedElements =
					specificationElement.getContainedElements();
			if(containedElements.size() == 1) {
				// Add reference from the element to this, so the lookups works as expected
				manager.getTreeNodeLookupTable().put(specificationElement, node);
				// Skip the specification element to get a more useful tree structure
				recursivlyCollectMetrics(node, containedElements.get(0), recursionLevel);
			} else {
				recursivlyCollectMetrics(node, specificationElement, recursionLevel);
			}
			// Add reference from the element to this, so the lookups works as expected
			manager.getTreeNodeLookupTable().put(currentElement, node);
		} else {

			// Iterate over all children and merge values
			var nodeDoubles = node.getStoredDoubles();
			var nodeIntegers = node.getStoredIntegers();
			for(IHierarchicElement containedElement : currentElement.getContainedElements()) {
				MetricTreeNode child = new MetricTreeNode();
				node.getChildren().add(child);
				recursivlyCollectMetrics(child, containedElement, recursionLevel + 1);

				for(MetricKey key : MetricKey.getCollectorKeys()) {
					var childDoubles = child.getStoredDoubles();
					var childIntegers = child.getStoredIntegers();
					if(childDoubles.containsKey(key)) {
						nodeDoubles.merge(key, childDoubles.get(key), Double::sum);
					}
					if(childIntegers.containsKey(key)) {
						nodeIntegers.merge(key, childIntegers.get(key), Integer::sum);
					}
				}
			}

			manager.getTreeNodeLookupTable().put(currentElement, node);

			GraphMetricsProvider.calculateBetweennessCentrality(currentElement, manager, false);
			GraphMetricsProvider.calculateBetweennessCentrality(currentElement, manager, true);
			node.getStoredDoubles().put(MetricKey.CLUSTERING_COEFFICIENT,
					GraphMetricsProvider.calculateClusteringCoefficent(currentElement));
		}
	}

	/**
	 * Collect all metrics provided by any {@link IMetricProvider}} for the given object and saves
	 * them in the {@link MetricTreeNode}.
	 * 
	 * @param node
	 *            to which the data is collected
	 * @param object
	 *            object to collect the data from
	 */
	@SuppressWarnings("unchecked")
	private void collectMetrics(MetricTreeNode node, EObject object) {

		List<IMetricProvider<? extends EObject>> providers = getAllMetricProviders(object);

		for(IMetricProvider<? extends EObject> provider : providers) {

			// The provider in the list has to have a compatible method for the actual type
			// as we just collected providers with a compatible type
			((IMetricProvider<EObject>)provider).collectMetrics(node, object);
		}
	}

	/** Get all suitable {@link IMetricProvider} for the given input. */
	private List<IMetricProvider<? extends EObject>> getAllMetricProviders(EObject input) {
		List<IMetricProvider<?>> providers = new ArrayList<>();
		for(Entry<Class<?>, List<IMetricProvider<?>>> migEntry : handlerMap.entrySet()) {
			if(migEntry.getKey().isAssignableFrom(input.getClass())) {
				providers.addAll(migEntry.getValue());
			}
		}
		return providers;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Metric Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void topLevelElementLoaded(ITopLevelElement element) {
		// Schedule metric collection when loading projects
		scheduleMetricCollection(element);
	}

	/** {@inheritDoc} */
	@Override
	public void topLevelElementAdded(ITopLevelElement element) {
		// Nothing done
	}

	/** {@inheritDoc} */
	@Override
	public void topLevelElementRemoved(ITopLevelElement element) {
		// Nothing done
	}

	/** {@inheritDoc} */
	@Override
	public void topLevelElementContentChanged(ITopLevelElement element) {
		scheduleMetricCollection(element);
	}

	/** @return the {@link MetricDataManager} for this service instance. */
	public MetricDataManager getMetricDataManagerInstance() {
		return metricDataManagerInstance;
	}
}
