/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.storage;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.fortiss.tooling.ext.quality.data.DataRootElement;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.kernel.model.IProjectRootElement;

/**
 * Allows the export of metrics to a csv file.
 * 
 * @author groh
 */
public class CSVFileWriter {

	/**
	 * Method to write the data into the default csv.
	 * <p>
	 * It will always write to the same file, if it is already present it will attempt to read the
	 * keys and convert them into {@link MetricKey} and then write these keys in the ordering that
	 * is already present. This is done such that the log file always remains consistent, but can
	 * cause newly added {@link MetricKey} not to be exported if they were not present when the file
	 * was exported the first time.
	 * 
	 * @param map
	 *            to convert into csv
	 * 
	 */
	public static void metricExtractionToCSV(Map<IProjectRootElement, DataRootElement> map) {

		// check if there is any data to export
		if(map == null || map.isEmpty()) {
			return;
		}

		// Ensure the folder exists
		if(!ModelQualityStorageManager.MODEL_QUALITY_PROJECT_DIR.exists()) {
			ModelQualityStorageManager.MODEL_QUALITY_PROJECT_DIR.mkdirs();
		}

		// path to csv file
		Path path = new File(ModelQualityStorageManager.MODEL_QUALITY_PROJECT_DIR,
				ModelQualityStorageManager.CSV_FILE_NAME).toPath();

		List<String> allKeys = new ArrayList<>();
		boolean createNewIndex = true;

		if(Files.exists(path)) {
			// Read first line and try to extract the keys from the already existing file
			try(BufferedReader reader = new BufferedReader(new FileReader(path.toFile()))) {
				String line = reader.readLine();
				if(line != null) {
					allKeys = Arrays.asList(line.split(","));
					createNewIndex = false;
				}
			} catch(IOException e) {
				e.printStackTrace();
			}
		}
		writeMetricsToFile(map, allKeys, path, createNewIndex);
	}

	/**
	 * Function to write the specified metrics into a csv file.
	 * 
	 * 
	 * @param map
	 *            of the metrics which are written into the file
	 * @param allKeys
	 *            a list of keys which will be extracted from the provided map
	 * @param path
	 *            location of the file
	 * @param createNewIndex
	 *            if a index should be written before writing any values
	 */
	private static void writeMetricsToFile(Map<IProjectRootElement, DataRootElement> map,
			List<String> allKeys, Path path, boolean createNewIndex) {
		try(var writer = new BufferedWriter(new FileWriter(path.toFile(), true))) {

			if(createNewIndex) {
				// Create new index and write it into the first line of the file
				allKeys.add("timestamp");
				allKeys.add("project_name");
				allKeys.add("root_element_name");
				allKeys.add("name");
				allKeys.add("children");
				// Collect all MetricKeys, convert them to a string and append them to the allKeys
				// list
				allKeys.addAll(Arrays.stream(MetricKey.values()).map(MetricKey::toString)
						.map(String::toLowerCase).collect(Collectors.toList()));

				// Write first line
				writer.write(String.join(",", allKeys));
				writer.write("\n");
			}
			// Remove first 4 keys as they are not provided by the provider
			List<String> valueKeys = allKeys.subList(5, allKeys.size());

			LocalDateTime dateTime = LocalDateTime.now();
			DateTimeFormatter formatter = DateTimeFormatter.ISO_LOCAL_DATE_TIME;
			String formattedDateTime = dateTime.format(formatter);

			map.forEach((rootElement, dataRootElement) -> {
				// iterate over all project roots
				String rootName = rootElement.getClass().getSimpleName();

				var topLevelElement = dataRootElement.getTopLevelElement();
				String projectName = topLevelElement.getSaveableName();

				var rootNode = dataRootElement.getRootNode();

				rootNode.traverseTree(node -> {
					var integers = node.getStoredIntegers();
					var doubles = node.getStoredDoubles();
					var strings = node.getStoredStrings();
					// collect IDs of children
					String children_ids = node.getChildren().stream()
							.map(n -> n.getStoredIntegers().get(MetricKey.UNQIUE_ID))
							.map(id -> id.toString()).reduce((a, b) -> a + " " + b).orElse("");

					// These 4 data entries are just concatenated in front
					var startStream = Stream.of(formattedDateTime, projectName, rootName,
							node.getName(), children_ids);
					// Collect all values from provider in the correct order and combine
					var values = Stream.concat(startStream, valueKeys.stream()
							.map(String::toUpperCase).map(MetricKey::valueOf).map(k -> {
								if(k.isStringValue())
									return strings.get(k);
								if(k.isDoubleValue())
									return doubles.get(k);
								return integers.get(k);
							}).map(String::valueOf)).collect(Collectors.joining(","));

					try {
						// write values
						writer.write(values);
						writer.write("\n");
					} catch(IOException e) {
						e.printStackTrace();
					}
				});
			});
		} catch(IOException e) {
			e.printStackTrace();
		}
	}
}
