/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.ui.commands;

import static java.util.Collections.emptyList;
import static org.conqat.ide.commons.ui.ui.WorkbenchUtils.getActiveWorkbenchWindow;
import static org.eclipse.jface.dialogs.MessageDialog.openError;
import static org.eclipse.jface.window.Window.OK;
import static org.eclipse.ui.PlatformUI.getWorkbench;
import static org.fortiss.tooling.ext.reuse.ui.prototypes.PrototypeProvider.DEFAULT_ROOT_REUSE_ELEMENT_NAME;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.manageExternalReferencesOfReuseElement;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryViewUtils.raiseAndUpdateReuseLibraryView;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getAllLocalReuseLibraries;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.hasReusePossibility;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.addElementToLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.createAndAddReuseLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.createElementForLibrary;
import static org.fortiss.tooling.kernel.ui.service.IContextMenuService.TOP_MOST_MENU_SECTION_ID;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.ui.ToolingReuseUIActivator;
import org.fortiss.tooling.ext.reuse.ui.dialog.AddElementToLibDialog;
import org.fortiss.tooling.kernel.internal.PrototypeService;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.ui.extension.IContextMenuContributor;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;

/**
 * Context menu contributions to include element adding to {@link ReuseLibrary}.
 *
 * @author bergemann
 */
public class AddElementContextMenu implements IContextMenuContributor {

	/** The title of the dialog shell(s). */
	private final String shellTitle = "Add Element to Reuse Library";

	/** {@inheritDoc} */
	@Override
	public List<IContributionItem> getContributedItems(EObject selection, ContextMenuContextProvider contextProvider) {
		if (selection != null && canBeAdded(selection)) {
			AddToLibAction action = new AddToLibAction(selection);
			List<IContributionItem> contributionItems = new ArrayList<IContributionItem>();
			contributionItems.add(new ActionContributionItem(action));
			return contributionItems;
		}
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public String getMenuSectionID() {
		return TOP_MOST_MENU_SECTION_ID;
	}

	/** True if the given element can be added to {@link ReuseLibrary}. */
	protected boolean canBeAdded(EObject element) {
		return hasReusePossibility(element);
	}

	/** Add one given element to specified {@link ReuseLibrary}. */
	protected void addElement(ReuseLibrary library, EObject elementToAdd, String saveName) {
		if (library == null || elementToAdd == null || saveName == null) {
			return;
		}

		EObject newElementForLibrary = createElementForLibrary(elementToAdd, library, saveName);

		try {
			addElementToLibrary(newElementForLibrary, library);
		} catch (Exception e) {
			openError(getActiveWorkbenchWindow().getShell(), "Error in '" + shellTitle + "'", e.getMessage());
		}

		PrototypeService.getInstance().updatePrototypes();

		String currentActionType = "addToLibrary";
		manageExternalReferencesOfReuseElement(elementToAdd, currentActionType, shellTitle);
	}

	/** The action "add element to {@link ReuseLibrary}". */
	private class AddToLibAction extends Action {

		/** The selected element. */
		private EObject selectedElement;

		/** The selected {@link ReuseLibrary}. */
		private ReuseLibrary selectedLibrary;

		/** All names of current {@link ReuseLibrary}s. */
		private List<String> currentExistingLibraryNames;

		/** All names of current {@link ReuseLibrary} elements. */
		private List<String> currentExistingLibraryElementNames;

		/** Constructor. Takes the selected element as input. */
		AddToLibAction(EObject element) {
			super("Add to Reuse Library", ToolingReuseUIActivator.getImageDescriptor("icons/library_add.png"));
			selectedElement = element;
		}

		/** {@inheritDoc} */
		@Override
		public void run() {
			selectedLibrary = null;
			currentExistingLibraryNames = new ArrayList<String>();
			List<ReuseLibrary> libraryList = getAllLocalReuseLibraries();

			Boolean createNewLibrary;
			String creationText;
			if (libraryList != null && !libraryList.isEmpty()) {
				// update existing library names
				for (ReuseLibrary tmpLibrary : libraryList) {
					currentExistingLibraryNames.add(getReuseElementName(tmpLibrary));
				}

				// ask for library selection
				final Shell shell = getWorkbench().getActiveWorkbenchWindow().getShell();
				final AddElementToLibDialog librarySelectDialog = new AddElementToLibDialog(shell, shellTitle);
				librarySelectDialog.setLibraryList(libraryList);
				librarySelectDialog.open();

				if (!librarySelectDialog.goAhead()) {
					return;
				}
				selectedLibrary = librarySelectDialog.getLibrary();
				createNewLibrary = librarySelectDialog.createNew();
				creationText = "A new library will be created in which the element can be added.\nEnter a name for the library:\n ";
			} else {
				createNewLibrary = true;
				creationText = "No library exists yet to add the element. A new one will be created.\nEnter a name for the library:\n ";
			}

			if (createNewLibrary) {
				final InputDialog creationDialog = new InputDialog(Display.getCurrent().getActiveShell(), shellTitle,
						creationText, DEFAULT_ROOT_REUSE_ELEMENT_NAME, new NewLibNameValidator());
				if (creationDialog.open() != OK) {
					return;
				}
				ICommandStackService.getInstance().runAsCommand(selectedElement, () -> {
					selectedLibrary = createAndAddReuseLibrary(creationDialog.getValue());
				});
			}

			if (selectedLibrary == null) {
				return;
			}

			// update existing element names
			currentExistingLibraryElementNames = new ArrayList<String>();
			EList<EObject> elementList = selectedLibrary.getReuseElementList();
			for (EObject libraryElement : elementList) {
				currentExistingLibraryElementNames.add(getReuseElementName(libraryElement));
			}

			// ask for new element name (when saving element to library)
			final InputDialog addingDialog = new InputDialog(Display.getCurrent().getActiveShell(), shellTitle,
					"Enter the name with which the element should be saved in the library '"
							+ getReuseElementName(selectedLibrary) + "':\n ",
					getReuseElementName(selectedElement), new AddedNameValidator());
			if (addingDialog.open() != OK) {
				return;
			}
			String newName = addingDialog.getValue();

			// execute "adding" action
			ICommandStackService.getInstance().runAsCommand(selectedElement, () -> {
				addElement(selectedLibrary, selectedElement, newName);
			});

			raiseAndUpdateReuseLibraryView();
		}

		/** Validator for the new {@link ReuseLibrary} name. */
		private class NewLibNameValidator implements IInputValidator {

			/** {@inheritDoc} */
			@Override
			public String isValid(String newText) {
				if (newText.equals("")) {
					return "A name needs to be specified.";
				}
				for (String name : currentExistingLibraryNames) {
					if (newText.equals(name)) {
						return "A library with this name already exists.";
					}
				}
				return null;
			}
		}

		/** Validator for the new (element) name. */
		private class AddedNameValidator implements IInputValidator {

			/** {@inheritDoc} */
			@Override
			public String isValid(String newText) {
				if (newText.equals("")) {
					return "A name needs to be specified.";
				}
				for (String name : currentExistingLibraryElementNames) {
					if (newText.equals(name)) {
						return "This name already exists. Please use another.";
					}
				}
				return null;
			}
		}
	}
}
