/*--------------------------------------------------------------------------+
|                                                                          |
| Copyright 2022 fortiss GmbH						                       |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.variability.ui.editor;

import static org.eclipse.jface.viewers.AbstractTreeViewer.ALL_LEVELS;
import static org.eclipse.swt.SWT.FILL;
import static org.eclipse.swt.SWT.NONE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.conqat.ide.commons.ui.jface.TreeContentProviderBase;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.TreeViewerColumn;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IEditorPart;
import org.fortiss.tooling.base.ui.editor.TreeViewerEditorBase;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation;
import org.fortiss.tooling.kernel.ui.extension.base.ModelEditorBindingBase;
import org.fortiss.tooling.kernel.ui.presentation.ConstraintViolationAwareLabelProviderBase;
import org.fortiss.tooling.kernel.ui.presentation.ModelElementLabelProvider;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;
import org.fortiss.tooling.kernel.ui.util.EObjectSelectionUtils;
import org.fortiss.variability.model.IAlternativeVariationPoint;
import org.fortiss.variability.model.presence.IHasPresenceCondition;
import org.fortiss.variability.model.presence.PresenceCondition;

/**
 * Editor for {@link IAlternativeVariationPoint}.
 * 
 * @author bayha
 */
public class AlternativeVariationPointEditor
		extends TreeViewerEditorBase<IAlternativeVariationPoint<?>> {

	/**
	 * Creates the {@link TreeViewer} instance.
	 * 
	 * @param parent
	 *            The parent, in which the {@link TreeViewer} is supposed to be
	 *            embedded in.
	 * @return The new {@link TreeViewer}.
	 */
	protected TreeViewer createTreeViewer(Composite parent) {
		return new TreeViewer(parent);
	}

	/** {@inheritDoc} */
	@Override
	protected void setupTreeViewer(TreeViewer treeViewer) {
		treeViewer.getTree().setHeaderVisible(true);
		treeViewer.getTree().setLinesVisible(true);
		treeViewer.getTree().setLayoutData(new GridData(FILL, FILL, true, true, 1, 1));
		treeViewer.setAutoExpandLevel(ALL_LEVELS);

		treeViewer.setContentProvider(new TreeContentProviderBase() {
			@Override
			public Object[] getChildren(Object parentElement) {
				List<Object> ret = new ArrayList<Object>();
				if(parentElement instanceof IAlternativeVariationPoint<?>) {
					ret.addAll(((IAlternativeVariationPoint<?>)parentElement).getAlternatives());
				}
				return ret.toArray();
			}

			/** {@inheritDoc} */
			@Override
			public Object[] getElements(Object inputElement) {
				if(inputElement instanceof AlternativeVariationPointEditor) {
					IAlternativeVariationPoint<?> avp =
							((AlternativeVariationPointEditor)inputElement).getEditedObject();
					return avp.getAlternatives().toArray();
				}
				if(inputElement instanceof IAlternativeVariationPoint<?>) {
					return ((IAlternativeVariationPoint<?>)inputElement).getAlternatives()
							.toArray();
				}
				return getElements(inputElement);
			}
		});

		treeViewer.addDoubleClickListener(new IDoubleClickListener() {
			@Override
			public void doubleClick(DoubleClickEvent event) {
				if(event.getSelection() instanceof IStructuredSelection) {
					EObject element = EObjectSelectionUtils.getFirstElement(event.getSelection());
					if(element != null) {
						IModelEditorBindingService.getInstance().openInEditor(element);
					}
				}
			}
		});
		TreeViewerColumn variantCol = new TreeViewerColumn(treeViewer, NONE);
		variantCol.getColumn().setResizable(true);
		variantCol.getColumn().setText(getAlternativesColumnLabel());
		variantCol.getColumn().setWidth(400);
		variantCol.setLabelProvider(new ModelElementLabelProvider());
		variantCol.setEditingSupport(new AlternativesNamesEditingSupport(treeViewer));

		TreeViewerColumn restCol = new TreeViewerColumn(treeViewer, NONE);
		restCol.getColumn().setResizable(true);
		restCol.getColumn().setText(getPresenceConditionColumnLabel());
		restCol.getColumn().setWidth(400);
		restCol.setLabelProvider(new ConstraintViolationAwareLabelProviderBase() {
			@Override
			public String getText(Object element) {
				if(element instanceof IHasPresenceCondition) {
					PresenceCondition pc = ((IHasPresenceCondition)element).getPresenceCondition();
					if(pc != null) {
						return pc.getStringRepresentation();
					}
				}
				return "";
			}

			/** {@inheritDoc} */
			@Override
			protected Image getBaseImage(Object element) {
				return getFallbackImage(element).createImage();
			}

			/** {@inheritDoc} */
			@Override
			protected Collection<IConstraintViolation<?>> getViolationsForObject(Object object) {
				if(object instanceof IHasPresenceCondition) {
					return super.getViolationsForObject(
							((IHasPresenceCondition)object).getPresenceCondition());
				}
				return super.getViolationsForObject(object);
			}
		});
		restCol.setEditingSupport(
				new PresenceConditionEditingSupportBase(treeViewer, dbc, editedObject) {

					@Override
					protected boolean canEdit(Object element) {
						return true;
					}
				});

		getSite().setSelectionProvider(treeViewer);
	}

	/**
	 * Retrieves the label for the presence condition column.
	 * 
	 * @return the label for the presence condition column.
	 */
	protected String getPresenceConditionColumnLabel() {
		return "Presence Conditions";
	}

	/**
	 * Retrieves the label for the alternatives column.
	 * 
	 * @return the label for the alternatives column.
	 */
	protected String getAlternativesColumnLabel() {
		return "Alternatives";
	}

	/** Editor Binding for {@link IAlternativeVariationPoint}. */
	public static class AlternativeVariationPointEditorBinding
			extends ModelEditorBindingBase<IAlternativeVariationPoint<?>> {
		/** {@inheritDoc} */
		@Override
		public Class<? extends IEditorPart> getEditorClass() {
			return AlternativeVariationPointEditor.class;
		}

		/** {@inheritDoc} */
		@Override
		public Class<? extends IEditorPart> getEditorClass(IAlternativeVariationPoint<?> object) {
			return getEditorClass();
		}

		/** {@inheritDoc} */
		@Override
		public String getLabel(IAlternativeVariationPoint<?> object) {
			return object.getName();
		}
	}
}
