/*--------------------------------------------------------------------------+
|                                                                          |
| Copyright 2022 fortiss GmbH						                       |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.util;

import static com.microsoft.z3.Z3javaAPIWrapper.createAnd;
import static com.microsoft.z3.Z3javaAPIWrapper.createImplication;
import static com.microsoft.z3.Z3javaAPIWrapper.createNot;
import static com.microsoft.z3.Z3javaAPIWrapper.createXor;
import static org.fortiss.variability.model.features.CrossFeatureConstraintType.EXCLUDES;
import static org.fortiss.variability.model.features.CrossFeatureConstraintType.REQUIRES;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import org.eclipse.emf.common.util.EList;
import org.fortiss.variability.model.features.AbstractAlternativeFeature;
import org.fortiss.variability.model.features.AbstractCompositionalFeature;
import org.fortiss.variability.model.features.AbstractCrossFeatureConstraint;
import org.fortiss.variability.model.features.AbstractFeature;
import org.fortiss.variability.model.features.AbstractFeatureModel;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Context;
import com.microsoft.z3.Z3javaAPIWrapper;

/**
 * Translation of Feature models to Z3.
 * 
 * FIXME: Refactor to variability kernel!
 * 
 * @author bayha
 */
public class FeatureModelTransformationUtils {

	/**
	 * Translates the given {@link AbstractFeature} to the Z3 API.
	 * 
	 * @param feature
	 *            The {@link AbstractFeature} to translate.
	 * @param ctx
	 *            The Z3 context to use.
	 * @return A {@link BoolExpr} representing 'feature'.
	 */
	public static BoolExpr feature2Z3(AbstractFeature feature, Context ctx) {
		return feature2Z3(feature, ctx, false);
	}

	/**
	 * Translates the given {@link AbstractFeature} to the Z3 API.
	 * 
	 * @param feature
	 *            The {@link AbstractFeature} to translate.
	 * @param ctx
	 *            The Z3 context to use.
	 * @param sanatizeName
	 *            Indicates, whether the name of the 'feature' is supposed
	 *            to be adapted for better compatibility with z3 (see
	 *            sanitizeZ3LiteralName(...)).
	 * @return A {@link BoolExpr} representing 'feature'.
	 */
	public static BoolExpr feature2Z3(AbstractFeature feature, Context ctx, boolean sanatizeName) {
		return createBoolConst(ctx, feature.getName(), sanatizeName);
	}

	/**
	 * Translates the given {@link AbstractFeatureModel} to the Z3 API.
	 * 
	 * @param featureModel
	 *            The {@link AbstractFeatureModel} to translate.
	 * @param ctx
	 *            The Z3 context.
	 * @param expr2FeatureMap
	 *            A {@link Map}, to which the mappings BoolExpr->Feature
	 *            will be added.
	 * @return A {@link BoolExpr} representing 'featureModel'.
	 */
	public static BoolExpr featureModel2Z3(AbstractFeatureModel featureModel, Context ctx,
			Map<BoolExpr, AbstractFeature> expr2FeatureMap) {
		return featureModel2Z3(featureModel, ctx, expr2FeatureMap, false);
	}

	/**
	 * Translates the given {@link AbstractFeatureModel} to the Z3 API.
	 * 
	 * @param featureModel
	 *            The {@link AbstractFeatureModel} to translate.
	 * @param ctx
	 *            The Z3 context.
	 * @param expr2FeatureMap
	 *            A {@link Map}, to which the mappings
	 *            BoolExpr->Feature will be added.
	 * @param sanatizeLiteralNames
	 *            Indicates, whether names are supposed to be
	 *            adapted for better compatibility with z3 (see
	 *            sanitizeZ3LiteralName(...)).
	 * @return A {@link BoolExpr} representing 'featureModel'.
	 */
	public static BoolExpr featureModel2Z3(AbstractFeatureModel featureModel, Context ctx,
			Map<BoolExpr, AbstractFeature> expr2FeatureMap, boolean sanatizeLiteralNames) {
		Queue<AbstractFeature> worklist = new ArrayDeque<AbstractFeature>();
		List<BoolExpr> clauses = new ArrayList<BoolExpr>();

		EList<AbstractFeature> rootFeatures = featureModel.getSubFeatures();
		worklist.addAll(rootFeatures);

		worklist.stream().filter(f -> !f.isOptional())
				.map(f -> createBoolConst(ctx, f.getName(), sanatizeLiteralNames))
				.forEach(e -> clauses.add(e));

		while(worklist.size() > 0) {
			AbstractFeature currentFeature = worklist.poll();
			BoolExpr currentLiteral =
					createBoolConst(ctx, currentFeature.getName(), sanatizeLiteralNames);
			expr2FeatureMap.put(currentLiteral, currentFeature);

			if(currentFeature instanceof AbstractAlternativeFeature) {
				EList<AbstractFeature> alternatives =
						((AbstractAlternativeFeature)currentFeature).getAlternatives();
				List<BoolExpr> altExprs = new ArrayList<BoolExpr>();
				for(AbstractFeature alt : alternatives) {
					BoolExpr altLiteral = createBoolConst(ctx, alt.getName(), sanatizeLiteralNames);
					expr2FeatureMap.put(altLiteral, alt);
					altExprs.add(altLiteral);

					// A1 -> AF; A2 -> AF; ...
					clauses.add(createImplication(ctx, altLiteral, currentLiteral));
				}

				// AF -> (A1 XOR A2 ...)
				clauses.add(createImplication(ctx, currentLiteral,
						createXor(ctx, altExprs.toArray(new BoolExpr[0]))));

				worklist.addAll(alternatives);
			}

			if(currentFeature instanceof AbstractCompositionalFeature) {
				EList<AbstractFeature> subFeatures =
						((AbstractCompositionalFeature)currentFeature).getSubFeatures();
				for(AbstractFeature f : subFeatures) {

					// subF -> F
					BoolExpr literal = createBoolConst(ctx, f.getName(), sanatizeLiteralNames);
					expr2FeatureMap.put(literal, f);
					clauses.add(createImplication(ctx, literal, currentLiteral));

					if(!f.isOptional()) {
						// F -> subF
						clauses.add(createImplication(ctx, currentLiteral, literal));
					}

					worklist.add(f);
				}
			}

			// Cross Tree Constraints.
			EList<AbstractCrossFeatureConstraint> constraints = currentFeature.getConstraints();
			for(AbstractCrossFeatureConstraint cfc : constraints) {
				AbstractFeature target = cfc.getTarget();
				if(target == null) {
					continue;
				}

				BoolExpr targetLiteral =
						createBoolConst(ctx, target.getName(), sanatizeLiteralNames);
				if(cfc.getType() == EXCLUDES) {
					clauses.add(
							createImplication(ctx, currentLiteral, createNot(ctx, targetLiteral)));
				}
				if(cfc.getType() == REQUIRES) {
					clauses.add(createImplication(ctx, currentLiteral, targetLiteral));
				}
			}
		}

		return createAnd(ctx, clauses.toArray(new BoolExpr[0]));
	}

	/**
	 * Creates a bool constant with the given name for the given ctx.
	 * 
	 * @param ctx
	 *            The {@link Context}.
	 * @param name
	 *            The name for the constant
	 * @param sanatizeName
	 *            Whether the name shall be adapted to a syntax, that can
	 *            be understood by z3.
	 * @return The {@link BoolExpr} constant.
	 */
	private static BoolExpr createBoolConst(Context ctx, String name, boolean sanatizeName) {
		if(sanatizeName) {
			name = sanitizeZ3LiteralName(name);
		}

		return Z3javaAPIWrapper.createBoolConst(ctx, name);
	}

	/**
	 * Adapts 'literal' to use only z3 conform characters.
	 * 
	 * @param literal
	 *            The {@link String} to sanitize.
	 * @return The z3 conform {@link String}.
	 */
	public static String sanitizeZ3LiteralName(String literal) {
		// remove whitespace from the literal or Z3 will break
		literal = literal.replaceAll("\\s+", "");
		// Replace umlaute
		literal = literal.replace("ä", "ae");
		literal = literal.replace("ü", "ue");
		literal = literal.replace("ö", "oe");
		literal = literal.replace("ß", "ss");
		// add the prefix "lit" because of possible initial number characters in the
		// name
		literal = "lit" + literal;
		return literal;
	}
}
